/***********************************/
/*                                 */
/* COMPILATEUR DE MINI C POUR MIPS */
/*                                 */
/*    Jacques Farr - ESSI         */
/*                                 */
/*         fvrier 2004            */
/*                                 */
/***********************************/

/****************************************************
  C'est une grammaire qui ressemble  celle de C, 
  mais ce n'est pas une vraie grammaire de C
****************************************************/

%{

#include <stdio.h>
#include <iostream>
using namespace std;

#include "arbre.h"

/* dfinis dans Lex */
extern int yylex (void);
extern FILE* yyin;

static GenreVariable genre_courant;
static const TypeSimple* type_base;

static ListeDeTypes* parametres_courants;

%}

%union {
  int              	ve;
  const char*           st;
  NoeudInstr*	        in;
  NoeudExpr*	        ex;
  NoeudDesig*     	ds;
  ListeArguments*	ar;
  ListeDeVariables*	va;
  ListeInstructions*	le;
  struct {
    Type*         typ;
    const char*   var;
    Fonction*     fct;
    ListeDeTypes* par;
  } de;
  struct {
    const Type* typ;
    Fonction*   fct;
  } di;
  struct { // pour sauvegarder les variables globales
    const TypeSimple* typ;
    Fonction*         fct;
    ListeDeTypes*     par;
    GenreVariable     gen;
  } fo;
}

%type<di> declaration_initiale declarateur
%type<de> pdecl tdecl sdecl
%type<ve> dim
%type<fo> ident_fonc 
%type<st> ident_var
%type<le> liste_instr corps
%type<in> instruction instruction_simple si tantque repeter retour bloc
%type<ex> condition expression facteur primaire
%type<ar> arguments
%type<va> variables
%type<ds> designation denotation

%token GOTO IF ELSE WHILE DO BREAK RETURN INT VOID PRINT READ

%token<ve> CONST
%token<st> IDENT STRING

%right '='
%left OU
%left ET
%nonassoc EQU NEQ GEQ LEQ '<' '>'
%left SLL SLR
%left '+' '-'
%left '*' '/' '%'
%right '!'
%right PREFIXE
%left '['

%%
progr
  : declarations
  ;

declarations
  : declaration
  | declarations declaration
  ;

declaration
  : declarations_simples
  | definition_fonction
  ;

definition_fonction
  : declaration_fonction '{' corps '}'	{ fonction_courante->DonnerCorps (new CorpsFonction (new NoeudBloc ($3)));
					  fonction_courante = 0; genre_courant = vglobale;
					} 
  ;

declaration_fonction
  : declaration_initiale		{ if ($1.fct == 0) yyerror ("pas une declaration de fonction"); 
					  genre_courant = vlocale; fonction_courante = $1.fct;
					}
  ;

declarations_simples
  : declaration_initiale suite_decl ';'	{ }
  ;

suite_decl
  : /* vide */				{ }
  | ',' declarateur suite_decl		{ }
  ;

declaration_initiale
  : type_initial declarateur		{ $$ = $2; }
  ;

type_initial
  : INT					{ type_base = &TypeSimple::entier; }
  | VOID				{ type_base = &TypeSimple::vide; }
  ;

declarateur 
  : pdecl				{ const Type* ty = type_base->Completer ($1.typ); 
					  $$.fct = $1.fct;
				          if ($1.fct != 0) {
					    TypeFonction* sig = new TypeFonction (ty, $1.par);
					    $1.fct->DonnerSignature (*sig);
					    $$.typ = sig;
					  }
					  else {
					    if (ty->Egal (TypeSimple::vide))
					      yyerror ("pas de variables/paramtre de type void");
					    if ($1.var != 0) {
					      if (fonction_courante != 0) {
						if (genre_courant == vparametre)
						  fonction_courante->DeclarerParametre ($1.var, ty);
						else
						  fonction_courante->DeclarerVariable ($1.var, ty);
					      }
					      else if (genre_courant == vglobale)
						VariableGlobale::Declarer ($1.var, ty);
					    }
					    $$.typ = ty;
				          }
					}
  ;

pdecl
  : '*' pdecl				{ $$.typ = new TypePointeur ($2.typ);
					  $$.var = $2.var; $$.fct = $2.fct; $$.par = $2.par; 
					}
  | tdecl				{ $$ = $1; }
  ;

tdecl
  : sdecl				{ $$ = $1; }
  | tdecl dim				{ int d = $2;
				          if ($1.typ == 0) { /* premire dimension */	
					    if (genre_courant == vparametre || $1.fct != 0)
					      d = 0; /* tableau indirect */
					  }
					  else if (d < 0) {
            				    yyerror ("dimension obligatoire pour variable");
					    d = 1;
					  }
					  $$.typ = new TypeTableau ($1.typ, d);
					  $$.var = $1.var; $$.fct = $1.fct; $$.par = $1.par; 
					}
  ;

dim
  : '[' ']'				{ $$ = -1; }
  | '[' CONST ']'			{ if ($2 <= 0) {
                                            yyerror ("pas de dimension <= 0"); $$ = 1;
					  } 
					  else
					    $$ = $2;
					}
  ;

sdecl
  : /* vide */ 				{ $$.var = 0; $$.fct = 0; $$.par = 0; $$.typ = 0;
					  if (genre_courant != vparametre)
					    yyerror ("pas de declaration anonyme, sauf les paramtres");
                                        }
  | ident_var			        { $$.var = $1; $$.fct = 0; $$.par = 0; $$.typ = 0; }
  | ident_fonc '(' parametres ')'       { /* on interdit donc syntaxiquement les pointeurs */
                                          /* sur fonction et les tableaux de fonctions     */
					  $$.var = 0; $$.fct = fonction_courante;
					  $$.par = parametres_courants; $$.typ = 0;
				          /* restituer les variables globales */
				          fonction_courante = $1.fct; parametres_courants = $1.par;
				          type_base = $1.typ; genre_courant = $1.gen;
					} 
  | '(' pdecl ')'			{ $$ = $2; }
  ;

ident_var
  : IDENT			        { $$ = $1; }
  ;

ident_fonc
  : IDENT				{ $$.fct = fonction_courante; 
					  $$.par = parametres_courants; 
				          $$.typ = type_base;
				          $$.gen = genre_courant;
					  if (genre_courant == vparametre) {
                                            yyerror ("pas de fonctions en parametre");
                                            /* on ne declarera rien : */
					    fonction_courante = 0; parametres_courants = 0;
					  }
					  else {
	  				    fonction_courante = &Fonction::Declarer ($1);
					    parametres_courants = new ListeDeTypes ();
                                          }
				          genre_courant = vparametre;
					} 
  ;

parametres 
  : /* vide */
  | liste_parametres
  ;

liste_parametres
  : declaration_parametre			{ }
  | liste_parametres ',' declaration_parametre	{ }
  ;

declaration_parametre
  : declaration_initiale		{ if (parametres_courants && $1.typ) 
					    parametres_courants->Append ($1.typ); 
					}
  ;

corps
  : decl_locales liste_instr		{ $$ = $2; }
  | liste_instr				{ $$ = $1; }
  | decl_locales			{ $$ = 0;  }
  ;

decl_locales
  : declarations_simples
  | decl_locales declarations_simples
  ;

liste_instr
  : instruction				{ if ($1 != 0) {
				            $$ = new ListeInstructions(); $$->push_back ($1);
					  }
					  else
					    $$ = 0;
					}
  | liste_instr instruction		{ if ($2 != 0) { 
				            if ($1 != 0) $$ = $1; else $$ = new ListeInstructions();
					    $$->push_back ($2);
					  }
					  else
					    $$ = $1;
					}
  ;

instruction 
  : bloc				{ $$ = $1; }
  | si					{ $$ = $1; }
  | tantque				{ $$ = $1; }
  | repeter ';'				{ $$ = $1; }
  | instruction_simple ';'	        { $$ = $1; }
  | expression ';'			{ $$ = new NoeudExprInstr ($1); }
  | error ';'		                { yyerrok; $$ = 0; }
  | ';'       	                	{ $$ = 0; }
  ;

instruction_simple
  : retour				{ $$ = $1; }
  | BREAK 				{ if (niveau_boucle == 0) yyerror ("le break n'est pas dans une boucle");
					  $$ = new NoeudSortie; 
					}
  | PRINT '(' arguments ')' 		{ $$ = new NoeudAffichage ($3); }
  | READ '(' variables ')'		{ $$ = new NoeudLecture ($3); }
  ;

si
  : IF condition instruction			{ if ($3)
						    $$ = new NoeudSi ($2, $3);
						  else
						    $$ = new NoeudExprInstr ($2);
						}
  | IF condition instruction ELSE instruction 	{ if ($3) 
						    $$ = new NoeudSi ($2, $3, $5);
						  else if ($5)
						    $$ = new NoeudSi (new NoeudNon ($2), $5);
						  else
						    $$ = new NoeudExprInstr ($2);
						}
  ;	

tantque
  : while condition instruction		{ $$ = new NoeudBoucle ($2, $3); --niveau_boucle; }
  ;

while
  : WHILE 				{ ++niveau_boucle; }
  ;

repeter
  : do bloc WHILE condition 	 	{ $$ = new NoeudBoucle ($2, $4); --niveau_boucle; }
  ;

do
  : DO	 				{ ++niveau_boucle; }
  ; 

retour
  : RETURN 				{ $$ = new NoeudRetour (0); }
  | RETURN expression 			{ $$ = new NoeudRetour ($2); }
  ;

bloc
  : '{' liste_instr '}'       		{ $$ = $2 ? new NoeudBloc ($2) : 0; }
  | '{' '}' 				{ $$ = 0; }
  ;

condition
  : '(' expression ')'			{ $$ = $2; }
  ;

expression
  : designation '=' expression  	{ $$ = new NoeudAffectation ($1, $3); }
  | expression OU expression		{ $$ = new NoeudOu ($1, $3); }
  | expression ET expression		{ $$ = new NoeudEt ($1, $3); }
  | expression EQU expression		{ $$ = new NoeudComparaison ($1, Operateurs::egal, $3); }
  | expression NEQ expression		{ $$ = new NoeudComparaison ($1, Operateurs::nonegal, $3); }
  | expression '<' expression		{ $$ = new NoeudComparaison ($1, Operateurs::infer, $3); }
  | expression LEQ expression		{ $$ = new NoeudComparaison ($1, Operateurs::infegal, $3); }
  | expression '>' expression		{ $$ = new NoeudComparaison ($1, Operateurs::super, $3); }
  | expression GEQ expression		{ $$ = new NoeudComparaison ($1, Operateurs::supegal, $3); }
  | expression SLL expression		{ $$ = new NoeudArithBinaire ($1, Operateurs::dec_gauche, $3); }
  | expression SLR expression		{ $$ = new NoeudArithBinaire ($1, Operateurs::dec_droit, $3); }
  | expression '+' expression		{ $$ = new NoeudArithBinaire ($1, Operateurs::plus, $3); }
  | expression '-' expression  		{ $$ = new NoeudArithBinaire ($1, Operateurs::moins, $3); }
  | expression '*' expression		{ $$ = new NoeudArithBinaire ($1, Operateurs::mult, $3); }
  | expression '/' expression		{ $$ = new NoeudArithBinaire ($1, Operateurs::divise, $3); }
  | expression '%' expression		{ $$ = new NoeudArithBinaire ($1, Operateurs::reste, $3); }
  | facteur				{ $$ = $1; }
  ;

facteur
  : CONST				{ $$ = new NoeudNombreEntier ($1); }
  | '-' CONST        %prec PREFIXE	{ $$ = new NoeudNombreEntier (-$2); }
  | STRING                              { $$ = new NoeudChaine (Chaine::Declarer($1)); }
  | '&' designation  %prec PREFIXE	{ $$ = new NoeudAdresse ($2); }
  | '!' facteur				{ $$ = new NoeudNon ($2); }
  | '-' primaire     %prec PREFIXE	{ $$ = new NoeudArithUnaire (Operateurs::moins_unaire, $2); }
  | primaire				{ /* la valeur d'un tableau est son adresse */
					  NoeudDesig* x = dynamic_cast<NoeudDesig*>($1);
					  const TypeTableau* t = x ? dynamic_cast<const TypeTableau*>(&x->type) : 0;
                                          $$ = t != 0 ? new NoeudAdresse (x, t->Element()) : $1;
					}
  ;

primaire
  : designation				{ $$ = $1; }
  | IDENT '(' ')'			{ $$ = new NoeudAppel (Fonction::Chercher ($1), 0); }
  | IDENT '(' arguments ')'		{ $$ = new NoeudAppel (Fonction::Chercher ($1), $3); }
  | '(' expression ')'			{ $2->Parentheser (); $$ = $2; }
  ;

designation
  : denotation				{ $$ = $1; }
  | '*' primaire %prec PREFIXE		{ $$ = new NoeudDerep ($2); }
  ;

denotation
  : IDENT				{ $$ = new NoeudVariable (Variable::Chercher ($1)); }
  | primaire '[' expression ']'		{ NoeudDesig* x = dynamic_cast<NoeudDesig*>($1);
					  if (!x) yyerror ("on ne peut indexer que des designations");
			                  $$ = new NoeudIndex (x, $3);
					}
  ;

arguments
  : expression				{ $$ = new ListeArguments(); $$->push_back ($1); }
  | arguments ',' expression		{ $1->push_back ($3); $$ = $1; }
  ;
  
variables
  : designation				{ $$ = new ListeDeVariables(); $$->push_back ($1); }
  | variables ',' designation		{ $1->push_back ($3); $$ = $1; }
  ;
  
%%

//------------------------

void yyerror (char* msg) { 
  const char* symb = yytext [0] == '\n' ? "fin de ligne"
                   : yytext [0] == '\0' ? "fin de fichier"
	           : yytext;
  Erreur (msg, yylineno, symb);
}

//------------------------

int yyparse ();

bool analyser_source (const char* p) {
  int l = strlen (p);
  char* f = new  char [l + 3];
  strcpy (f, p); strcpy (f + l, ".c");
  yyin = fopen (f, "r");

  if (yyin == 0) {
    cerr << "impossible d'ouvrir " << f << '\n';
    return false;
  }

  fichier = f;
  erreurs = false;
  genre_courant = vglobale;
  yyparse ();
  return !erreurs;
}
