
%token IDENTIFIER "Identifier"
%token PUBLIC "public"
%token PROTECTED "protected"
%token PRIVATE "private"
%token STATIC "static"
%token ABSTRACT "abstract"
%token FINAL "final"
%token NATIVE "native"
%token SYNCHRONIZED "synchronized"
%token TRANSIENT "transient"
%token VOLATILE "volatile"
%token Type VariableInitializer ThrowsOpt MethodBody ConstructorBody
%token FormalParameterListOpt

%start ClassBodyDeclaration

%%

SimpleTypeName:
  "Identifier" ;

/* 19.8.1 Productions from 8.1: Class Declaration */

ClassBodyDeclaration:
  ClassMemberDeclaration |
  ConstructorDeclaration ;

ClassMemberDeclaration:
  FieldDeclaration |
  MethodDeclaration ;

/* 19.8.2 Productions from 8.3: Field Declarations */

FieldDeclaration:
  FieldModifiersOpt Type VariableDeclarators ';' ;

VariableDeclarators:
  VariableDeclarator |
  VariableDeclarators ',' VariableDeclarator ;

VariableDeclarator:
  VariableDeclaratorId |
  VariableDeclaratorId '=' VariableInitializer ;

VariableDeclaratorId:
  "Identifier" ;

FieldModifiersOpt:
  /* empty */ |
  FieldModifiers ;

FieldModifiers:
  FieldModifier |
  FieldModifiers FieldModifier ;

FieldModifier:
  "public" | "protected" | "private" |
  "final" | "static" | "transient" | "volatile" ;

/* 19.8.3 Productions from 8.4: Method Declarations */

MethodDeclaration:
  MethodHeader MethodBody ;


MethodHeader:
  MethodModifiersOpt ResultType MethodDeclarator ThrowsOpt ;

ResultType:
  Type |
  "void" ;

MethodDeclarator:
  "Identifier" '(' FormalParameterListOpt ')' ;

MethodModifiersOpt:
  /* empty */ |
  MethodModifiers;

MethodModifiers:
  MethodModifier |
  MethodModifiers MethodModifier ;

MethodModifier:
  "public" | "protected" | "private" |
  "abstract" | "static" | "final" | "synchronized" | "native" ;

/* 19.8.5 Productions from 8.6: Constructor Declarations */

ConstructorDeclaration:
  ConstructorModifiersOpt ConstructorDeclarator ThrowsOpt ConstructorBody ;

ConstructorDeclarator:
  SimpleTypeName '(' FormalParameterListOpt ')' ;

ConstructorModifiersOpt:
  /* empty */ |
  ConstructorModifiers ;

ConstructorModifiers:
  ConstructorModifier |
  ConstructorModifiers ConstructorModifier ;

ConstructorModifier:
  "public" | "protected" | "private" ;
